-- CSEEE4830 - TerrorMouse
-- by Ron Weiss

--basic digital waveguide. . .
library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_ARITH.all;
use IEEE.STD_LOGIC_UNSIGNED.all;

entity waveguide is
  port (
    clk        : in  std_logic := '0';
    reset      : in  std_logic := '0';
    enable     : in  std_logic := '0';
    -- length of delay line (max 256 samples => min freq of 93 Hz)
    length     : in  std_logic_vector(7 downto 0):= "11111111";   
    sample_out : out std_logic_vector(15 downto 0)
    );
end waveguide;

architecture karplus_strong of waveguide is
  component delayline 
    port (
      clk      : in  std_logic;
      we       : in  std_logic;                -- write enable
      en       : in  std_logic;                -- read enable
      addr     : in  std_logic_vector(7 downto 0);
      data_in  : in  std_logic_vector(15 downto 0);
      data_out : out std_logic_vector(15 downto 0)
    );
  end component;
  
  signal cnt             : std_logic_vector(10 downto 0):="00000000000";
  
  signal curr_sample     : std_logic_vector(15 downto 0):="0000000000000000";

  -- delayline RAM control signals
  signal we       : std_logic := '0';
  signal en       : std_logic := '0';
  signal addr     : std_logic_vector(7 downto 0)  := "00000000";
  signal data_in  : std_logic_vector(15 downto 0) := "0000000000000000";
  signal data_out : std_logic_vector(15 downto 0) := "0000000000000000";
  
  -- signals to hold the state of the delay line
  signal start_dl         : std_logic:='0';
  signal read_data        : std_logic:='0'; 
  signal write_data       : std_logic:='0';
  signal read_data2       : std_logic:='0'; 
  signal write_data2      : std_logic:='0';    
  signal feedback_sample  : std_logic:='0';
  signal feedback_sample2 : std_logic:='0';
  signal feedback_sample3 : std_logic:='0';
  
  -- save last two signals in delay line:
  signal y_k_1           : std_logic_vector(15 downto 0):="0000000000000000";
  signal y_k             : std_logic_vector(15 downto 0):="0000000000000000";

  signal delay_clk       : std_logic:='0';
  signal resetting       : std_logic:='0';
  
begin  
  process (clk)
  begin
    if clk'event and clk = '1' then     
      cnt <= cnt + 1;
    end if;
  end process;

 
  -- this is where the magic happens. yay state machines
  process (clk, reset, start_dl)
  begin
    if clk'event and clk='1' then
       if enable = '1' then  
         if reset='1' then 
           -- if reset = 1 reset the contents of the delay line
           resetting <= '1';
           addr    <= "00000000";
           curr_sample <= "0000000000000000";
           read_data <= '0';
           write_data<= '0';
           feedback_sample <= '0';
         elsif resetting = '1' then
           en   <= '1';
           we   <= '1';
         
           --some random numbers to set the initial state of the delay line
           if addr = 0 then
             data_in <= "1111111111100110";
           elsif addr = 1 then
             data_in <= "0000000001000010";
           elsif addr = 2 then
             data_in <= "0000000000111100";
           elsif addr = 4 then
             data_in <= "0000100000100100";
           elsif addr = 5 then
             data_in <= "1111111000100001";
           elsif addr = 6 then
             data_in <= "1111111000100111";
           elsif addr = 7 then
             data_in <= "0111111000100001";
           elsif addr = 10 then
             data_in <= "0000000010000000";
           elsif addr = 15 then
             data_in <= "0000010010001100";
           elsif addr = 17 then
             data_in <= "1111000010000000";
           elsif addr = 20 then
             data_in <= "1111000010000000";
           elsif addr = 30 then
             data_in <= "1111000000000000";
           elsif addr = 50 then
             data_in <= "1111000010000000";
           else
             data_in <= "0000000000000000";
           end if;
           
           addr  <= addr + 1;
        
           if addr = length then
             resetting <= '0';
             start_dl  <= '1';
           end if;
         elsif start_dl = '1' or cnt = 0 then
           addr <= length - 1;        
           read_data <= '1';
           start_dl  <= '0';
           we <= '0';
         elsif read_data= '1' then
           en   <= '1';
           we   <= '0';          
         
           read_data  <= '0';
           read_data2  <= '1';	   
         elsif read_data2 = '1' then
           read_data2<='0';
           write_data<='1';          
           addr <= addr + 1;
         elsif write_data = '1' then
           en   <= '1';
           we   <= '1';

           data_in <= data_out;

           if addr = length - 1 then  
             y_k_1 <= data_out;
           end if;                       

           write_data <= '0';
           write_data2<= '1';        
         elsif write_data2 = '1' then
           we <= '0';
           write_data2 <= '0';
           -- is this the end of the delay line?
           if addr = 0 then
             feedback_sample <= '1';
             addr <= "00000000";
           else
             read_data <= '1';
             addr <= addr - 2;  
           end if;
         elsif feedback_sample = '1' then
           en   <= '1';
           we   <= '0';
          
           addr <= length;
         
           feedback_sample  <= '0'; 	 
           feedback_sample2 <= '1';
         elsif feedback_sample2 = '1' then
           feedback_sample2 <= '0';
           feedback_sample3 <= '1';
         elsif feedback_sample3 = '1' then
           y_k <= data_out;
           
           -- loop filter:
           -- ks loop filter H(z) = -.5*(1-z^-1)
           -- karplus strong loop filter: out = .-5*(y(k) + y(k-1))
           curr_sample <= (('1' & (not data_out(15 downto 1)
                                   + not y_k_1(15 downto 1))) + 1);
         
           -- data to write into delay line (same as curr_sample)
           data_in     <= (('1' & (not data_out(15 downto 1)
                                   + not y_k_1(15 downto 1))) + 1);
           
           -- loop filter designed in DSP:
           --         0.8995 + 0.1087z^-1
           -- Hl(z) = -------------------
           --           1 + 0.0136z^-1
            
           -- feed curr_sample back to beginning of delay line
           en   <= '1';
           we   <= '1';
           addr <= "00000000";
           
           feedback_sample3  <= '0';
         else
           en <= '0';
         end if;
       else
         curr_sample <= "0000000000000000";
       end if;
    end if;
  end process;
 
  sample_out(15 downto 0) <= curr_sample;
  
  dl : delayline
    port map (
      clk      => clk,
      we       => we,
      en       => en,
      addr     => addr,
      data_in  => data_in,
      data_out => data_out
    );
  
end karplus_strong;

